/*
 * safefile package    http://www.cs.wisc.edu/~kupsch/safefile
 *
 * Copyright 2007-2008, 2011 James A. Kupsch
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include "safe_open.h"
#include "safe_fopen.h"



int main()
{
    const char* fn = "test_file";
    const char* linkFn = "test_file.link";

    int r;	/* result codes		*/
    int f;	/* file descriptor	*/

    FILE *F;

    mode_t allMask = 0777;
    mode_t userMask = 0600;
    mode_t myUmask = 0;

    struct stat curStat;
    struct stat prevStat;

    unsigned long curMask;
    unsigned long correctMask;

    const char data[] = "0123456789";
    ssize_t n;

    /* no umask	*/
    umask(myUmask);


    /*************************************************************************
     * remove file if it exists
     */

    r = unlink(fn);
    if (r == -1 && errno != ENOENT)  {
	perror("unlink");
	exit(1);
    }


    /*************************************************************************
     * test safe_create_fail_if_exists with no prior file
     */

    f = safe_create_fail_if_exists(fn, O_WRONLY, allMask);
    if (f == -1)  {
	perror("safe_create_fail_if_exists");
	exit(2);
    }

    r = fstat(f, &curStat);

    if (r == -1)  {
	perror("stat");
	exit(3);
    }

    curMask = curStat.st_mode & allMask;
    correctMask = allMask & ~myUmask;
    if (curMask != correctMask)  {
	fprintf(stderr, "bad permissions (%04lo) not %04lo\n",
		    curMask, correctMask);
	exit(4);
    }

    prevStat = curStat;

    r = close(f);
    if (r == -1)  {
	perror("close");
	exit(5);
    }


    /*************************************************************************
     * test safe_create_fail_if_exists with prior file
     */

    f = safe_create_fail_if_exists(fn, O_WRONLY, allMask);
    if (f != -1)  {
	fputs("safe_create_fail_if_exists success and it should not", stderr);
	exit(10);
    }


    /*************************************************************************
     * test safe_create_replace_file_if_exists with prior file
     */

    r = link(fn, linkFn);
    if (r == -1)  {
	perror("link");
	exit(20);
    }

    f = safe_create_replace_if_exists(fn, O_WRONLY, userMask);
    if (f == -1)  {
	perror("safe_create_replace_if_exists");
	exit(21);
    }

    n = write(f, data, sizeof data);
    if (n != sizeof data)  {
	if (n == -1)  {
	    perror("write");
	    exit(22);
	}  else  {
	    fputs("partial write", stderr);
	    exit(23);
	}
    }

    r = fstat(f, &curStat);

    if (r == -1)  {
	perror("stat");
	exit(24);
    }

    if (curStat.st_dev == prevStat.st_dev && curStat.st_ino == prevStat.st_ino)  {
	fputs("inode not different\n", stderr);
	exit(25);
    }

    /* perms should have changed */
    curMask = curStat.st_mode & allMask;
    correctMask = userMask & ~myUmask;
    if (curMask != correctMask)  {
	fprintf(stderr, "bad permissions (%04lo) not %04lo\n",
		    curMask, correctMask);
	exit(26);
    }

    prevStat = curStat;

    r = close(f);
    if (r == -1)  {
	perror("close");
	exit(25);
    }

    r = unlink(linkFn);
    if (r == -1)  {
	perror("unlink");
	exit(26);
    }


    /*************************************************************************
     * test_safe_create_keep_if_exists with prior file
     */

    f = safe_create_keep_if_exists(fn, O_WRONLY, allMask);
    if (f == -1)  {
	perror("safe_create_keep_if_exists");
	exit(30);
    }

    r = fstat(f, &curStat);

    if (r == -1)  {
	perror("stat");
	exit(31);
    }

    if (curStat.st_dev != prevStat.st_dev || curStat.st_ino != prevStat.st_ino)  {
	fputs("inode different\n", stderr);
	exit(32);
    }

    if (curStat.st_size != prevStat.st_size)  {
	fputs("file contents are changed", stderr);
	exit(33);
    }

    /* perms should not have changed */
    curMask = curStat.st_mode & allMask;
    correctMask = userMask & ~myUmask;
    if (curMask != correctMask)  {
	fprintf(stderr, "bad permissions (%04lo) not %04lo\n",
		    curMask, correctMask);
	exit(34);
    }

    prevStat = curStat;

    r = close(f);
    if (r == -1)  {
	perror("close");
	exit(35);
    }


    /*************************************************************************
     * test_safe_open_no_create with prior file
     */

    f = safe_open_no_create(fn, O_RDONLY);
    if (f == -1)  {
	perror("safe_open_no_create");
	exit(40);
    }

    r = fstat(f, &curStat);

    if (r == -1)  {
	perror("stat");
	exit(41);
    }

    if (curStat.st_dev != prevStat.st_dev || curStat.st_ino != prevStat.st_ino)  {
	fputs("inode different\n", stderr);
	exit(42);
    }

    if (curStat.st_size != prevStat.st_size)  {
	fputs("file contents are changed", stderr);
	exit(43);
    }

    /* perms should not have changed */
    curMask = curStat.st_mode & allMask;
    correctMask = userMask & ~myUmask;
    if (curMask != correctMask)  {
	fprintf(stderr, "bad permissions (%04lo) not %04lo\n",
		    curMask, correctMask);
	exit(44);
    }

    prevStat = curStat;

    r = close(f);
    if (r == -1)  {
	perror("close");
	exit(45);
    }


    /*************************************************************************
     * test_safe_create_replace_if_exists with no prior file
     */

    r = unlink(fn);
    if (r == -1)  {
	perror("unlink");
	exit(60);
    }

    f = safe_create_keep_if_exists(fn, O_WRONLY, userMask);
    if (f == -1)  {
	perror("safe_create_replace_if_exists");
	exit(61);
    }

    r = fstat(f, &curStat);

    if (r == -1)  {
	perror("stat");
	exit(62);
    }

    /* perms should not have changed */
    curMask = curStat.st_mode & allMask;
    correctMask = userMask & ~myUmask;
    if (curMask != correctMask)  {
	fprintf(stderr, "bad permissions (%04lo) not %04lo\n",
		    curMask, correctMask);
	exit(63);
    }

    prevStat = curStat;

    r = close(f);
    if (r == -1)  {
	perror("close");
	exit(64);
    }


    /*************************************************************************
     * test_safe_create_keep_if_exists with no prior file
     */

    r = unlink(fn);
    if (r == -1)  {
	perror("unlink");
	exit(70);
    }

    f = safe_create_keep_if_exists(fn, O_WRONLY, userMask);
    if (f == -1)  {
	perror("safe_create_keep_if_exists");
	exit(71);
    }

    r = fstat(f, &curStat);

    if (r == -1)  {
	perror("stat");
	exit(72);
    }

    /* perms should not have changed */
    curMask = curStat.st_mode & allMask;
    correctMask = userMask & ~myUmask;
    if (curMask != correctMask)  {
	fprintf(stderr, "bad permissions (%04lo) not %04lo\n",
		    curMask, correctMask);
	exit(73);
    }

    prevStat = curStat;

    r = close(f);
    if (r == -1)  {
	perror("close");
	exit(74);
    }


    /*************************************************************************
     * test_safe_open_no_create with no prior file
     */

    r = unlink(fn);
    if (r == -1)  {
	perror("unlink");
	exit(80);
    }

    f = safe_open_no_create(fn, O_RDONLY);
    if (f != -1 && errno != ENOENT)  {
	if (f)  {
	    fputs("safe_open_no_create", stderr);
	    exit(81);
	}  else  {
	    perror("safe_open_no_create should have failed with ENOENT");
	    exit(82);
	}
    }


    /*************************************************************************
     * test_safe_open_no_create with O_CREAT flag, should fail
     */

    f = safe_open_no_create(fn, O_RDONLY | O_CREAT);
    if (f != -1)  {
	fputs("safe_open_no_create", stderr);
	exit(80);
    }

    if (errno != EINVAL)  {
	perror("safe_open_no_create should have failed with EINVAL");
	exit(81);
    }


    /*************************************************************************
     * final cleanup
     */

    (void)unlink(fn);
    (void)unlink(linkFn);


    /*************************************************************************
     * test fopen on /dev/null
     */
    
    F = safe_fopen_wrapper_follow("/dev/null", "w", 0666);
    if (F == NULL)  {
	perror("safe_fopen_wrapper failed");
	exit(90);
    }  else  {
	fclose(F);
    }


    /*************************************************************************
     * success
     */

    return 0;
}
